/*
  data.csvを読み込み構造体に保存しクラスタリングを行う
  
  データ構造
  1. 1人分のデータを覚える構造体を宣言
  　構造体の内容
  　　番号
  　　据置プレイ時間
  　　携帯プレイ時間
  　　最長時間
  　　所属クラスタ
  2. 各クラスタ間の距離は2次元配列で表す．
  　　統合されて消滅したクラスタと他のクラスタの距離は-1とする．

  アルゴリズム
  1. csvファイルから構造体へのデータの読み込み，
  　　クラスタの初期化
  2. クラスタ間の距離の初期化
  3. 最小距離のクラスタを求める
  5. クラスタ数がK以下になるまで5.と6.を繰り返す関数
  4. 最小距離のクラスタの結合
  6. クラスタ間の距離表示関数
  7. 構造体にコピーしたデータを表示
 */

#include<stdio.h>

#define SIZE 85 /* 扱うデータの人数 */
//#define K 10

/* 1人分のデータを記憶する構造体 */
struct data{
  int number; /* 個人番号 */
  char brain; /* 脳波パターン */
  int handheld; /* 携帯型ゲームプレイ時間 */
  int video; /* 据置型ゲームプレイ時間 */
  int longest; /* 最長プレイ時間 */
  int cluster; /* 所属クラスタ */
};

/* プロトタイプ宣言 */

/* csvファイルのデータを構造体配列に読み込む関数 */
void read(struct data DATA[]);

/* 2乗関数 */
int square(int number);

/* 2人分のデータの平方ユークリッド距離を求める関数 */
int distance(struct data data1, struct data data2);

/* 各クラスタ間の初期化関数 */
void initialize(struct data DATA[], int DISTANCE[][SIZE]);

/* 最小距離のクラスタを求める関数 */
void minimum_cluster(struct data DATA[], int DISTANCE[][SIZE], int *min1, int *min2);

/* クラスタを結合する関数 */
void combine_cluster(struct data DATA[], int DISTANCE[][SIZE], int min1, int min2);

/* 構造体配列の各個人データの出力関数 */
void print(struct data DATA[]);

/* クラスタ間距離出力関数 */
void print_cluster_distance(int DISTANCE[][SIZE]);

/* クラスタ出力関数 */
void print_cluster(struct data DATA[], int k);

int main(int argc, char *argv[]){
  struct data DATA[SIZE]; /* 個人データを記憶する構造体の配列 */
  int DISTANCE[SIZE][SIZE]; /* 各クラスタ間の距離を記憶する配列 */
  int min1, min2; /* 距離が最小となるクラスタのペアを記憶 */
  int i, k;

  k=atoi(argv[1]);

  /* csvファイルのデータを構造体配列に読み込む関数 */
  read(DATA);

  /* 各クラスタ間の距離の初期化関数 */
  initialize(DATA, DISTANCE);
  
  for(i=SIZE; i>k; i--){
    /* 最小距離のクラスタを求める関数 */
    minimum_cluster(DATA, DISTANCE, &min1, &min2);
    
    /* クラスタを結合する関数 */
    combine_cluster(DATA, DISTANCE, min1, min2);
  }

  /* クラスタ間距離出力関数 */
  //print_cluster_distance(DISTANCE);
  
  /* 構造体配列の各個人データの出力関数 */
  //print(DATA);

  /* クラスタ出力関数 */
  print_cluster(DATA, k);

  return 0;
}

/* csvファイルのデータを構造体配列に読み込む関数 */
void read(struct data DATA[]){
  int num = 0; /* 読み込んでいる数字の記憶 */
  int state = 0; /* 何列目のセルか記憶 */
  int i = 0; /* 何人目の個人データか記憶 */
  char ch; /* 現在読み込んでいる文字の記憶 */
  char brain = '\0'; /* 脳波 */
  FILE *input; /* 入力用ファイルポインタ */
  
  input = fopen("data.csv", "rt"); /* ファイルを開く */

  /* ファイルが終了するまで1文字ずつ読み込む */
  while(fscanf(input,"%c",&ch)!=EOF){    
    if(ch>='0' && ch<='9') /* 数字読み込み */
      /* これまで読み込んだ値を10倍し現在読み込み中の値を加える */
      /*  ch - '0' は数字の文字コードとの差の調整 */
      num = num * 10 + ch - '0';
    else if(ch>='A' && ch<='Z') /* 英字（脳波）読み込み */
      brain  = ch;
    else if(ch=='"') ; /* '"' 読み込み */
    else if(ch==';'){ /* ; 読み込み，; で列の区切りを認識 */
      /* これまで読み込んだ数字や英字の代入 */
      if(state==0) /* 1列目読み込み */
	DATA[i].number = DATA[i].cluster = num - 1; /* 個人番号 */
      if(state==1) /* 2列目読み込み */
	DATA[i].brain = brain; /* 脳波パターン */
      if(state==2) /* 3列目読み込み */
	DATA[i].handheld = num; /* 携帯型ゲーム */
      if(state==3) /* 4列目読み込み */
	DATA[i].video = num; /* 据置型ゲーム */
      num = 0; /* 次読み込む数値を0で初期化 */
      state++; /* 次の列に移る事を記憶 */
    }
    else{ /* 改行読み込み，5列目読み込み */
      DATA[i].longest = num; /* これまで読み込んだ数値は最長プレイ時間 */
      num = 0; /* 次読み込む数値を0で初期化 */
      state = 0; /* 次の行の1列目に移る事を記憶 */
      brain = '\0';
      i++; /* 次の個人データに記憶することを記憶 */
    }
  }
}

/* 2乗関数 */
int square(int number){
  return number * number;
}

/* 2人分のデータの平方ユークリッド距離を求める関数 */
int distance(struct data data1, struct data data2){
  return square(data1.handheld - data2.handheld) + square(data1.video - data2.video) + square(data1.longest - data2.longest);
}

/* 各クラスタ間の距離の初期化関数 */
void initialize(struct data DATA[], int DISTANCE[][SIZE]){
  int i, j;  
  for(i=0; i<SIZE; i++)
    for(j=i+1; j<SIZE; j++)
      DISTANCE[i][j] = distance(DATA[i],DATA[j]);
}

/* 最小距離のクラスタを求める関数 */
void minimum_cluster(struct data DATA[], int DISTANCE[][SIZE], int *min1, int *min2){
  int i, j;
  *min1 = *min2 = -1;
  for(i=0; i<SIZE; i++)
    for(j=i+1; j<SIZE; j++)
      if(DISTANCE[i][j]!=-1 && (*min1==-1 && *min2==-1 || DISTANCE[i][j]<DISTANCE[*min1][*min2])){
	  *min1 = i;
	  *min2 = j;
	}
  //printf("minimum distance cluseters %3d and %3d, distance: %d\n", DATA[*min1].cluster, DATA[*min2].cluster, DISTANCE[*min1][*min2]);
  *min1 = DATA[*min1].cluster;
  *min2 = DATA[*min2].cluster;
}

/* クラスタを結合する関数 */
void combine_cluster(struct data DATA[], int DISTANCE[][SIZE], int min1, int min2){
  int i, j;
  for(i=0; i<SIZE; i++)
    if(DATA[i].cluster==min2){
      //printf("number %3d: cluster %3d => %3d\n", DATA[i].number, DATA[i].cluster, min1);
      DATA[i].cluster = min1;
      for(j=0; j<i; j++){
	if(DISTANCE[j][i]!=-1 && DISTANCE[j][i]<DISTANCE[j][min1])
	  DISTANCE[j][min1]=DISTANCE[j][i];
	DISTANCE[j][i] = -1;
	//	printf("DISTANCE[%3d][%3d]=-1\n", j, i);
      }
      for(j=i+1; j<SIZE; j++){
	if(DISTANCE[i][j]!=-1 && DISTANCE[i][j]<DISTANCE[min1][j])
	  DISTANCE[min1][j]=DISTANCE[i][j];
	DISTANCE[i][j] = -1;
	//	printf("DISTANCE[%3d][%3d]=\n", i, j);
      }
    }
  return;
}

/* 構造体配列の各個人データの出力関数
   個人番号，携帯型プレイ時間，据置型プレイ時間，最長プレイ時間，所属クラスタ
   の表示
 */
void print(struct data DATA[]){
  int i;
  printf("num b hnd vid lng cls\n");
  for(i=0; i<SIZE; i++)
    if(DATA[i].brain!=0)
      printf("%3d %c %3d %3d %3d %3d\n", DATA[i].number, DATA[i].brain, DATA[i].handheld, DATA[i].video, DATA[i].longest, DATA[i].cluster);
    else
      printf("%3d   %3d %3d %3d %3d\n", DATA[i].number, DATA[i].handheld, DATA[i].video, DATA[i].longest, DATA[i].cluster);
}

/* クラスタ間距離出力関数 */
void print_cluster_distance(int DISTANCE[][SIZE]){
  int i, j;
  for(i=0; i<SIZE; i++)
    for(j=i+1; j<SIZE; j++)
      if(DISTANCE[i][j]!=-1)
	printf("distance of cluster %3d and %3d is %d\n", i, j, DISTANCE[i][j]);
}

/* クラスタ出力関数 */
void print_cluster(struct data DATA[], int k){
  int i, j, size, total, G, H, N, cluster[k];
  for(j=0; j<k; j++) cluster[j]=-1;
  for(i=0; i<SIZE; i++)
    for(j=0;j<k; j++)
      if(cluster[j]==DATA[i].cluster) break;
      else if(cluster[j]==-1){
	cluster[j]=DATA[i].cluster;
	break;
      }
  for(j=0; j<k; j++){
    size = total = G = H = N = 0;
    for(i=0; i<SIZE; i++){
      if(DATA[i].cluster==cluster[j]){
	size++;
	total += DATA[i].handheld + DATA[i].video + DATA[i].longest;
	switch(DATA[i].brain){
	case 'G': G++; break;
	case 'H': H++; break;
	case 'N': N++; break;
	}
      }
    }
    printf("cluster number: %3d, size: %2d, average time: %3d, G: %2d(%3d%%), H: %2d(%3d%%), N: %2d(%3d%%)\n", cluster[j], size, total/(3*size), G, 100*G/(G+H+N), H, 100*H/(G+H+N), N, 100*N/(G+H+N));
  }
}
